#version 450 core

in vec2 fragTexCoordinate;

out vec4 color;

uniform sampler2D colorTexture;
uniform sampler2D depthTexture;
uniform sampler2D normalTexture;
uniform bool enableContours;
uniform bool showNormals;
uniform float near;
uniform float far;

vec3 edgeDetectionNormals();
vec3 edgeDetectionDepth();

void main() {
    if (!enableContours) {
        if (showNormals) {
            color = texture(normalTexture, fragTexCoordinate);
        } else {
            color = texture(colorTexture, fragTexCoordinate);
        }
        return;
    }

    // If the normal is (1, 1, 1) (= Skybox), we skip contours because the depth texture has some glitches for the skybox
    // Other normals from other object which would go the same direction as (1, 1, 1) are always normalized, so this
    // scenario is exclsive to the skybox
    if (texture(normalTexture, fragTexCoordinate) == vec4(1.0f)) {
        color = texture(colorTexture, fragTexCoordinate);
        return;
    }

    vec3 edgeCol = edgeDetectionNormals();
    edgeCol = max(edgeCol, edgeDetectionDepth());
    if (length(edgeCol) == 0) {
        if (showNormals) {
            color = texture(normalTexture, fragTexCoordinate);
        } else {
            color = texture(colorTexture, fragTexCoordinate);
        }
    } else {
        color = vec4(0, 0, 0, 1);
    }
}

vec3 edgeDetectionNormals() {
    vec2 offset = 1.0f / textureSize(normalTexture, 0);
    vec2 offsets[9] = vec2[](
        vec2(-offset.x,  offset.y), // top-left
        vec2( 0.0f,    offset.y), // top-center
        vec2( offset.x,  offset.y), // top-right
        vec2(-offset.x,  0.0f),   // center-left
        vec2( 0.0f,    0.0f),   // center-center
        vec2( offset.x,  0.0f),   // center-right
        vec2(-offset.x, -offset.y), // bottom-left
        vec2( 0.0f,   -offset.y), // bottom-center
        vec2( offset.x, -offset.y)  // bottom-right
    );

    float kernel[9] = float[](
        -1, -1, -1,
        -1, 8, -1,
        -1, -1, -1
    );

    vec3 sampleTex[9];
    for (int i = 0; i < 9; i++) {
        sampleTex[i] = vec3(texture(normalTexture, fragTexCoordinate.st + offsets[i]));
    }
    vec3 col = vec3(0.0);
    for (int i = 0; i < 9; i++) {
        col += sampleTex[i] * kernel[i];
    }

    if (length(col) < 0.1) {
        return vec3(0, 0, 0);
    } else {
        return vec3(1, 1, 1);
    }
}

vec3 edgeDetectionDepth() {
    vec2 offset = 1.0f / textureSize(depthTexture, 0);
    vec2 offsets[9] = vec2[](
        vec2(-offset.x,  offset.y), // top-left
        vec2( 0.0f,    offset.y), // top-center
        vec2( offset.x,  offset.y), // top-right
        vec2(-offset.x,  0.0f),   // center-left
        vec2( 0.0f,    0.0f),   // center-center
        vec2( offset.x,  0.0f),   // center-right
        vec2(-offset.x, -offset.y), // bottom-left
        vec2( 0.0f,   -offset.y), // bottom-center
        vec2( offset.x, -offset.y)  // bottom-right
    );

    float kernel[9] = float[](
        -1, -1, -1,
        -1, 8, -1,
        -1, -1, -1
    );

    vec3 sampleTex[9];
    for (int i = 0; i < 9; i++) {
        sampleTex[i] = vec3(texture(depthTexture, fragTexCoordinate.st + offsets[i]));
        sampleTex[i] = near * far / (far + sampleTex[i] * (near - far));
    }
    vec3 col = vec3(0.0);
    for (int i = 0; i < 9; i++) {
        col += sampleTex[i] * kernel[i];
    }

    if (length(col) < 0.1f) {
        return vec3(0, 0, 0);
    } else {
        return vec3(1, 1, 1);
    }
}